<?php

class XenForo_Importer_myvBulletin4 extends XenForo_Importer_vBulletin
{
	public static function getName()
	{
		return 'vBulletin 4.x';
	}
    
    public function getSteps() 
    {
        return array_merge(parent::getSteps(), array(
            'postThanks' => array(
                'title' => 'Import Post Thanks',
                'depends' => array('users', 'threads')
            ),
	        'postLikes' => array(
	                        'title' => 'Import Post Likes',
	                        'depends' => array('users', 'threads')
	        ),
	        'social' => array(
	        				'title' => 'Import Social Groups',
	        				'depends' => array('threads')
	        ),
	        'sgimages' => array(
	        				'title' => 'Import Social Group Images',
	        				'depends' => array('social')
	        ),
	        'albums' => array(
	        				'title' => 'Import Albums',
	        				'depends' => array('threads')
	        ),
	        'blogs' => array(
	        				'title' => 'Import Blogs',
	        				'depends' => array('threads')
	        ),
	        'articles' => array(
	        				'title' => 'Import CMS Articles',
	        				'depends' => array('threads')
	        ),
        	'attachments' => array(
        				'title' => new XenForo_Phrase('import_attached_files'),
        				'depends' => array('threads')
        ),
        ));
    }

    public function stepAttachments($start, array $options)
    {
        $options = array_merge(array(
            'path' => isset($this->_config['attachmentPath']) ? trim($this->_config['attachmentPath']) : '',
            'limit' => 50,
            'processed' => 0,
            'max' => false
        ), $options);

        $sDb = $this->_sourceDb;
        $prefix = $this->_prefix;
        $model = $this->_importModel;

        if ($options['max'] === false)
        {
            $content = $sDb->fetchAll('
                SELECT contenttypeid, class
                FROM ' . $prefix . 'contenttype
                WHERE class IN (\'Post\',\'BlogEntry\',\'Article\')
            ');

            $temp = Array('0');

            foreach ($content AS $type)
            {
                $temp[] = $type['contenttypeid'];
                $options['attach'.$type['class']] = $type['contenttypeid'];
            }

            $options['allTypes'] = implode(',',$temp);

            $data = $sDb->fetchRow('
                SELECT MAX(attachmentid) AS max, COUNT(attachmentid) AS rows
                FROM ' . $prefix . 'attachment
                WHERE state = \'visible\'
                AND contenttypeid IN ('.$options['allTypes'].')
            ');

            $options = array_merge($options,$data);
        }

        $attachments = $sDb->fetchAll($sDb->limit('
                SELECT attachmentid, attachment.userid, attachment.dateline, filedataid, 
                filename, counter, contentid as postid, filedata.userid as fileuserid, contenttypeid as typeid
                FROM ' . $prefix . 'attachment AS attachment
                INNER JOIN ' . $prefix . 'filedata AS filedata USING (filedataid)
                WHERE attachmentid > ' . $sDb->quote($start) . '
                    AND state = \'visible\'
                    AND contenttypeid IN ('.$options['allTypes'].')
                ORDER BY attachmentid
            ', $options['limit']
        ));

        if (!$attachments)
        {
            return true;
        }

        $next = 0;
        $total = 0;

        $cmsIdMap = $this->getCmsIdsMapFromArray($attachments, 'postid');
        $blogIdMap = $this->getBlogIdsMapFromArray($attachments, 'postid');
        $postIdMap = $model->getPostIdsMapFromArray($attachments, 'postid');

        $allPostids = array_unique(array_merge($postIdMap,$blogIdMap,$cmsIdMap)); // Keys are reset !
        $posts = $model->getModelFromCache('XenForo_Model_Post')->getPostsByIds($allPostids);

        $userIdMap = $model->getUserIdsMapFromArray($attachments, 'userid');

        foreach ($attachments AS $attachment)
        {
            $next = $attachment['attachmentid'];

            $attachment['userid'] = $this->_mapLookUp($userIdMap, $attachment['userid'],0);

            if ($attachment['typeid'] == $options['attachPost'])
            { 
                $attachment['postid'] = $this->_mapLookUp($postIdMap, $attachment['postid'],0);
            }

            if (isset($options['attachArticle']) AND $attachment['typeid'] == $options['attachArticle'])
            { 
                $attachment['postid'] = $this->_mapLookUp($cmsIdMap, $attachment['postid'],0);
            }

            if (isset($options['attachBlogEntry']) AND $attachment['typeid'] == $options['attachBlogEntry'])
            { 
                $attachment['postid'] = $this->_mapLookUp($blogIdMap, $attachment['postid'],0);
            }

            if (!$attachment['postid'])
            {
                continue;
            }
            else
            {
                $attachment['message'] = &$posts[$attachment['postid']]['message'];
            }

            $success = $this->_makeAttachment($options, $attachment);

            if ($success)
            {
                $total++;
            }
        }

        $options['processed'] += $total; 
        $this->_session->incrementStepImportTotal($total);

        return array($next, $options, $this->_getProgressOutput($options['processed'], $options['rows']));
    }
    
    protected function getCmsIdsMapFromArray(array $source, $key)
    {
        $model = $this->_importModel;
        $userIds = array();
        foreach ($source AS $data)
        {
            $userIds[] = $data[$key];
        }
        return $model->getImportContentMap('cmsfirst', $userIds);
    }

    protected function getBlogIdsMapFromArray(array $source, $key)
    {
        $model = $this->_importModel;
        $userIds = array();
        foreach ($source AS $data)
        {
            $userIds[] = $data[$key];
        }
        return $model->getImportContentMap('blogfirst', $userIds);
    }
    
    protected function _makeAttachment($options, &$attachment)
    {
        $sDb = $this->_sourceDb;
        $prefix = $this->_prefix;
        $model = $this->_importModel;

        if (!$options['path'])
        {
            $fData = $sDb->fetchOne('
                SELECT filedata
                FROM ' . $prefix . 'filedata
                WHERE filedataid = ' . $sDb->quote($attachment['filedataid'])
            );

            if ($fData === '')
            {
                return false;
            }

            $attachFile = tempnam(XenForo_Helper_File::getTempDir(), 'xf');

            if (!$attachFile || !@file_put_contents($attachFile, $fData))
            {
                return false;
            }

            $isTemp = true;
        }
        else
        {
            $attachFileOrig = "$options[path]/" . implode('/', str_split($attachment['fileuserid'])) . "/$attachment[filedataid].attach";

            if (!file_exists($attachFileOrig))
            {
                return false;
            }

            $attachFile = tempnam(XenForo_Helper_File::getTempDir(), 'xf');
            copy($attachFileOrig, $attachFile);

            $isTemp = true;
        }

        $success = $model->importPostAttachment(
            $attachment['attachmentid'],
            $this->_convertToUtf8($attachment['filename']),
            $attachFile,
            $attachment['userid'],
            $attachment['postid'],
            $attachment['dateline'],
            array('view_count' => $attachment['counter']),
            array($this, 'processAttachmentTags'),
            $attachment['message']
        );

        if ($isTemp)
        {
            @unlink($attachFile);
        }

        return $success;
    }
    
    public function stepPostThanks($start, array $options)
    {
        $options = array_merge(array(
            'limit' => 200,
            'max' => false
        ), $options);

        $sDb = $this->_sourceDb;
        $prefix = $this->_prefix;

        /* @var $model PostThanksImporter_Model_Import */
        $model = $this->_importModel;
        
        if ($options['max'] === false)
        {
            $options['max'] = $sDb->fetchOne('
                SELECT MAX(id)
                FROM ' . $prefix . 'post_thanks
            ');
        }

        $postThanks = $sDb->fetchAll($sDb->limit(
            '
                SELECT pt.*, post.userid as content_userid
                FROM ' . $prefix . 'post_thanks AS pt
                LEFT JOIN ' . $prefix . 'post AS post USING (postid)
                WHERE pt.id > ?
                ORDER BY pt.id
            ', $options['limit']
        ), $start);
        if (!$postThanks)
        {
            return true;
        }

        $next = 0;
        $total = 0;

        $userIds = array();
        $postIds = array();
        foreach ($postThanks AS $thanks)
        {
            $userIds[] = $thanks['userid'];
            $userIds[] = $thanks['content_userid'];
            $postIds[] = $thanks['postid'];
        }
        $userIdMap = $model->getImportContentMap('user', $userIds);
        $postIdMap = $model->getImportContentMap('post', $postIds);
        XenForo_Db::beginTransaction();

        foreach ($postThanks AS $thanks)
        {
            $next = $thanks['id'];

            $likeUserId = $this->_mapLookUp($userIdMap, $thanks['userid']);
            $contentUserId = $this->_mapLookUp($userIdMap, $thanks['content_userid']);
            $postId = $this->_mapLookUp($postIdMap, $thanks['postid']);
            if (!$likeUserId OR !$contentUserId OR !$postId)
            {
                continue;
            }
            
            $model->importPostThanks($postId, $contentUserId, $likeUserId, $thanks['date']);
            $total++;
        }


        XenForo_Db::commit();

        $this->_session->incrementStepImportTotal($total);
        
        // 1.0.0 Beta 2 introduces percentage progress format
        if (XenForo_Application::$versionId > 1000031)
        {
            return array($next, $options, $this->_getProgressOutput($next, $options['max']));
        }
        else 
        {
            return array($next, $options, "$next / $options[max]");
        }
    }
    
    public function stepPostLikes($start, array $options)
    {
    	$options = array_merge(array(
                'limit' => 200,
                'max' => false
    	), $options);
    
    	$sDb = $this->_sourceDb;
    	$prefix = $this->_prefix;
    
    	/* @var $model PostLikessImporter_Model_Import */
    	$model = $this->_importModel;
    
    	if ($options['max'] === false)
    	{
    		$options['max'] = $sDb->fetchOne('
                    SELECT MAX(helpfulanswerid)
                    FROM ' . $prefix . 'helpfulanswer
                ');
    	}
    
    	$postLikes = $sDb->fetchAll($sDb->limit(
                '
                    SELECT ha.*, post.userid as content_userid
                    FROM ' . $prefix . 'helpfulanswer AS ha
                    LEFT JOIN ' . $prefix . 'post AS post USING (postid)
                    WHERE ha.helpfulanswerid > ?
                    ORDER BY ha.helpfulanswerid
                ', $options['limit']
    	), $start);
    	if (!$postLikes)
    	{
    		return true;
    	}
    
    	$next = 0;
    	$total = 0;
    
    	$userIds = array();
    	$postIds = array();
    	foreach ($postLikes AS $likes)
    	{
    		$userIds[] = $likes['userid'];
    		$userIds[] = $likes['content_userid'];
    		$postIds[] = $likes['postid'];
    	}
    	$userIdMap = $model->getImportContentMap('user', $userIds);
    	$postIdMap = $model->getImportContentMap('post', $postIds);
    	XenForo_Db::beginTransaction();
    
    	foreach ($postLikes AS $likes)
    	{
    		$next = $likes['helpfulanswerid'];
    
    		$likeUserId = $this->_mapLookUp($userIdMap, $likes['userid']);
    		$contentUserId = $this->_mapLookUp($userIdMap, $likes['content_userid']);
    		$postId = $this->_mapLookUp($postIdMap, $likes['postid']);
    		if (!$likeUserId OR !$contentUserId OR !$postId)
    		{
    			continue;
    		}
    
    		$model->importPostThanks($postId, $contentUserId, $likeUserId, $likes['dateline']);
    		$total++;
    	}
    
    
    	XenForo_Db::commit();
    
    	$this->_session->incrementStepImportTotal($total);
    
    	// 1.0.0 Beta 2 introduces percentage progress format
    	if (XenForo_Application::$versionId > 1000031)
    	{
    		return array($next, $options, $this->_getProgressOutput($next, $options['max']));
    	}
    	else
    	{
    		return array($next, $options, "$next / $options[max]");
    	}
    }
    
    //
    
    public function stepAlbums($start, array $options)
    {
    	$options = array_merge(array(
    			'path' => isset($this->_config['attachmentPath']) ? trim($this->_config['attachmentPath']) : '',
    			'limit' => 50,
    			'alimit' => 4, // images per post
    			'processed' => 0,
    			'max' => false
    	), $options);
    
    	$sDb = $this->_sourceDb;
    	$prefix = $this->_prefix;
    	$model = $this->_importModel;
    
    	if ($options['max'] === false)
    	{
    		$content = $sDb->fetchOne('
    				SELECT contenttypeid
    				FROM ' . $prefix . 'contenttype
    				WHERE class = \'Album\'
    			');
    
    		$options['albumType'] = intval($content);
    
    		$data = $sDb->fetchRow('
    				SELECT MAX(attachmentid) AS max, COUNT(attachmentid) AS rows
    				FROM ' . $prefix . 'attachment
    				WHERE state = \'visible\'
    				AND contenttypeid = ' . $sDb->quote($options['albumType'])
    		);
    
    		$options = array_merge($options,$data);
    			
    		if ($options['max'])
    		{
    			// Get board name
    			$board = $sDb->fetchOne('
    					SELECT value
    					FROM ' . $prefix . 'setting
    					WHERE varname = \'bbtitle\'
    				');
    
    			XenForo_Db::beginTransaction();
    			$options['forumid'] = $this->buildImportForum($this->_convertToUtf8('Albums ('.$board.')', true), 940);
    			XenForo_Db::commit();
    		}
    
    		return array(0, $options, "Processing Albums ...");
    	}
    
    	$attachments = $sDb->fetchAll($sDb->limit('
    				SELECT attachmentid, album.userid, attachment.dateline, title, username,
    				filedataid, filename, counter, albumid, filedata.userid as fileuserid, album.state
    				FROM ' . $prefix . 'attachment AS attachment
    				INNER JOIN ' . $prefix . 'filedata AS filedata USING (filedataid)
    				INNER JOIN ' . $prefix . 'album AS album ON (album.albumid = attachment.contentid)
    				INNER JOIN ' . $prefix . 'user AS user ON (album.userid = user.userid)
    				WHERE albumid > ' . $sDb->quote($start) . '
    				AND attachment.state = \'visible\'
    				AND contenttypeid = ' . $sDb->quote($options['albumType']) . '
    				ORDER BY albumid, attachmentid
    			', $options['limit']
    	));
    
    	if (!$attachments)
    	{
    		// Rebuild the permission cache
    		XenForo_Model::create('XenForo_Model_Node')->updateNestedSetInfo();
    		XenForo_Model::create('XenForo_Model_Permission')->rebuildPermissionCache();
    		return true;
    	}
    
    	$next = 0;
    	$last = 0;
    	$total = 0;
    	$position = 0;
    	$userIdMap = $model->getUserIdsMapFromArray($attachments, 'userid');
    
    	foreach ($attachments AS $attachment)
    	{
    		$next = $attachment['albumid'];
    
    		if ($last <> $next)
    		{
    			$last = $next;
    			$userid = $model->mapUserId($attachment['userid'], 0);
    			$username = $this->_convertToUtf8($attachment['username'], true);
    			$username = $this->mbTrim($username,50,$userid);
    
    			$import = array(
    					'title' => $this->_convertToUtf8('Album : '.$attachment['title'], true),
    					'node_id' => $options['forumid'],
    					'user_id' => $userid,
    					'username' => $username,
    					'discussion_open' => 1,
    					'post_date' => $attachment['dateline'],
    					'reply_count' => 0,
    					'view_count' => 0,
    					'sticky' => 0,
    					'discussion_state' => $attachment['state'] == 'public' ? 'visible' : 'moderated',
    					'last_post_date' => $attachment['dateline'],
    					'last_post_username' => $username
    			);
    
    			$threadid = $model->importThread(0, $import);
    
    			$position = 0;
    			$acount = $options['alimit'];
    			$model->logImportData('album', $attachment['albumid'], $threadid);
    		}
    
    		if ($acount >= $options['alimit'])
    		{
    			if ($position != 0)
    			{
    				$this->_db->query('
    						UPDATE xf_thread
    						SET reply_count = reply_count + 1
    						WHERE thread_id = ?
    					', $threadid);
    			}
    
    			$acount = 0;
    			$position++;
    
    			$import = array(
    					'thread_id' => $threadid,
    					'user_id' => $userid,
    					'username' => $username,
    					'post_date' => $attachment['dateline']+$position,
    					'message' => "Album Images $position \n\n",
    					'attach_count' => 0,
    					'ip' => '0.0.0.0',
    					'position' => $position,
    					'message_state' => 'visible'
    			);
    
    			$postid = $model->importPost(0, $import);
    		}
    
    		$attachment['userid'] = $userid;
    		$attachment['postid'] = $postid;
    
    		unset($attachment['message']);
    		$success = $this->_makeAttachment($options, $attachment);
    
    		if ($success)
    		{
    			$total++;
    			$acount++;
    		}
    	}
    
    	$options['processed'] += $total;
    	$this->_session->incrementStepImportTotal($total);
    
    	return array($next, $options, $this->_getProgressOutput($options['processed'], $options['rows']));
    }
    
    public function stepSocial($start, array $options)
    {
    	$options = array_merge(array(
    			'max' => false,
    			'limit' => 100,
    			'processed' => 0,
    			'postLimit' => 500,
    			'postDateStart' => 0,
    	), $options);
    
    	$sDb = $this->_sourceDb;
    	$prefix = $this->_prefix;
    	$model = $this->_importModel;
    
    	if ($options['max'] === false)
    	{
    		$data = $sDb->fetchRow('
    				SELECT MAX(discussionid) AS max, COUNT(discussionid) AS rows
    				FROM ' . $prefix . 'discussion
    			');
    
    		$options = array_merge($options,$data);
    
    		if ($options['max'])
    		{
    			// Get board name
    			$board = $sDb->fetchOne('
    					SELECT value
    					FROM ' . $prefix . 'setting
    					WHERE varname = \'bbtitle\'
    				');
    
    			XenForo_Db::beginTransaction();
    			$options['forumid'] = $this->buildImportForum($this->_convertToUtf8('Social Groups ('.$board.')', true), 910);
    			XenForo_Db::commit();
    		}
    
    		return array(0, $options, "Processing Social Groups ...");
    	}
    
    	$threads = $sDb->fetchAll($sDb->limit('
    				SELECT discussion.*, groupmessage.*, socialgroup.name
    				FROM ' . $prefix . 'discussion AS discussion
    				INNER JOIN ' . $prefix . 'socialgroup AS socialgroup USING (groupid)
    				INNER JOIN ' . $prefix . 'groupmessage AS groupmessage 
    				ON (discussion.firstpostid = groupmessage.gmid)
    				WHERE discussion.discussionid >= ' . $sDb->quote($start) . '
    				ORDER BY discussion.discussionid
    			', $options['limit']
    	));
    
    	if (!$threads)
    	{
    		// Rebuild the permission cache
    		XenForo_Model::create('XenForo_Model_Node')->updateNestedSetInfo();
    		XenForo_Model::create('XenForo_Model_Permission')->rebuildPermissionCache();
    		return true; // All done.
    	}
    
    	$next = 0;
    	$total = 0;
    	$totalPosts = 0;
    
    	XenForo_Db::beginTransaction();
    
    	foreach ($threads AS $thread)
    	{
    		if (trim($thread['postusername']) === '' || trim($thread['title']) === '')
    		{
    			continue;
    		}
    
    		$postDateStart = $options['postDateStart'];
    
    		$options['postDateStart'] = 0;
    		$next = $thread['discussionid'] + 1;
    
    		$maxPosts = $options['postLimit'] - $totalPosts;
    
    		$posts = $sDb->fetchAll($sDb->limit('
    					SELECT groupmessage.* 
    					FROM ' . $prefix . 'groupmessage AS groupmessage
    					INNER JOIN ' . $prefix . 'user AS user 
    					ON (groupmessage.postuserid = user.userid)
    					WHERE groupmessage.discussionid = ' . $sDb->quote($thread['discussionid']) . '
    					AND groupmessage.dateline > ' . $sDb->quote($postDateStart) . '
    					ORDER BY groupmessage.dateline
    				', $maxPosts
    		));
    
    		if (!$posts)
    		{
    			if ($postDateStart)
    			{
    				$total++;
    			}
    			continue;
    		}
    
    		if ($postDateStart)
    		{
    			$threadId = $options['threadid'];
    			$position = $this->_db->fetchOne('
    					SELECT MAX(position)
    					FROM xf_post
    					WHERE thread_id = ?
    				', $threadId);
    		}
    		else
    		{
    			if (!$options['forumid'])
    			{
    				continue;
    			}
    
    			$import = array(
    					'title' => $this->_convertToUtf8($thread['name'], true).' : '.$this->_convertToUtf8($thread['title'], true),
    					'node_id' => $options['forumid'],
    					'user_id' => $model->mapUserId($thread['postuserid'], 0),
    					'username' => $this->_convertToUtf8($thread['postusername'], true),
    					'discussion_open' => 1,
    					'post_date' => $thread['dateline'],
    					'reply_count' => $thread['visible'] - 1,
    					'view_count' => 1,
    					'sticky' => 0,
    					'last_post_date' => $thread['lastpost'],
    					'last_post_username' => $this->_convertToUtf8($thread['lastposter'], true)
    			);
    
    			$import['username'] = $this->mbTrim($import['username'],50,$import['user_id']);
    
    			switch ($thread['state'])
    			{
    				case 'moderation': $import['discussion_state'] = 'moderated'; break;
    				default: $import['discussion_state'] = $thread['state'];
    			}
    
    			$threadId = $model->importThread(0, $import);
    
    			$options['threadid'] = $threadId;
    			$model->logImportData('discussion', $thread['discussionid'], $threadId);
    
    			if (!$threadId)
    			{
    				continue;
    			}
    
    			$position = -1;
    		}
    
    		if ($threadId)
    		{
    			$userIdMap = $model->getUserIdsMapFromArray($posts, 'postuserid');
    
    			foreach ($posts AS $post)
    			{
    				if (trim($post['postusername']) === '' || trim($post['pagetext']) === '')
    				{
    					continue;
    				}
    
    				$post['pagetext'] = $this->_convertVideo($post['pagetext']);
    
    				$import = array(
    						'thread_id' => $threadId,
    						'user_id' => $this->_mapLookUp($userIdMap, $post['postuserid'], 0),
    						'username' => $this->_convertToUtf8($post['postusername'], true),
    						'post_date' => $post['dateline'],
    						'message' => $this->_convertToUtf8($post['pagetext']),
    						'attach_count' => 0,
    						'ip' => $post['ipaddress']
    				);
    
    				$import['username'] = $this->mbTrim($import['username'],50,$import['user_id']);
    
    				switch ($post['state'])
    				{
    					case 'deleted': $import['message_state'] = 'deleted'; $import['position'] = $position; break;
    					case 'moderation': $import['message_state'] = 'moderated'; $import['position'] = $position; break;
    					default: $import['message_state'] = 'visible'; $import['position'] = ++$position;
    				}
    
    				$postid = $model->importPost(0, $import);
    				$model->logImportData('message', $post['gmid'], $postid);
    
    				$options['postDateStart'] = $post['dateline'];
    				$totalPosts++;
    			}
    		}
    
    		if (count($posts) < $maxPosts)
    		{
    			$total++;
    			$options['postDateStart'] = 0;
    		}
    		else
    		{
    			break;
    		}
    	}
    
    	XenForo_Db::commit();
    
    	if ($options['postDateStart'])
    	{
    		$next--;
    	}
    
    	$options['processed'] += $total;
    	$this->_session->incrementStepImportTotal($total);
    
    	return array($next, $options, $this->_getProgressOutput($options['processed'], $options['rows']));
    }
    
    public function stepSGImages($start, array $options)
    {
    	$options = array_merge(array(
    			'path' => isset($this->_config['attachmentPath']) ? trim($this->_config['attachmentPath']) : '',
    			'limit' => 50,
    			'alimit' => 4, // images per post
    			'processed' => 0,
    			'max' => false
    	), $options);
    
    	$sDb = $this->_sourceDb;
    	$prefix = $this->_prefix;
    	$model = $this->_importModel;
    
    	if ($options['max'] === false)
    	{
    		$content = $sDb->fetchOne('
    				SELECT contenttypeid
    				FROM ' . $prefix . 'contenttype
    				WHERE class = \'SocialGroup\'
    			');
    
    		$options['albumType'] = intval($content);
    
    		$data = $sDb->fetchRow('
    				SELECT MAX(attachmentid) AS max, COUNT(attachmentid) AS rows
    				FROM ' . $prefix . 'attachment
    				WHERE state = \'visible\'
    				AND contenttypeid = ' . $sDb->quote($options['albumType'])
    		);
    
    		$options = array_merge($options,$data);
    			
    		if ($options['max'])
    		{
    			// Get board name
    			$board = $sDb->fetchOne('
    					SELECT value
    					FROM ' . $prefix . 'setting
    					WHERE varname = \'bbtitle\'
    				');
    
    			XenForo_Db::beginTransaction();
    			$options['forumid'] = $this->buildImportForum($this->_convertToUtf8('Social Group Images ('.$board.')', true), 950);
    			XenForo_Db::commit();
    		}
    
    		return array(0, $options, "Processing SG Images ...");
    	}
    
    	$attachments = $sDb->fetchAll($sDb->limit('
    				SELECT attachmentid, creatoruserid as userid, attachment.dateline, name as title, username,
    				filedataid, filename, counter, groupid, filedata.userid as fileuserid, type as state
    				FROM ' . $prefix . 'attachment AS attachment
    				INNER JOIN ' . $prefix . 'filedata AS filedata USING (filedataid)
    				INNER JOIN ' . $prefix . 'socialgroup AS socialgroup ON (socialgroup.groupid = attachment.contentid)
    				INNER JOIN ' . $prefix . 'user AS user ON (socialgroup.creatoruserid = user.userid)
    				WHERE groupid > ' . $sDb->quote($start) . '
    				AND attachment.state = \'visible\'
    				AND contenttypeid = ' . $sDb->quote($options['albumType']) . '
    				ORDER BY groupid, attachmentid
    			', $options['limit']
    	));
    
    	if (!$attachments)
    	{
    		// Rebuild the permission cache
    		XenForo_Model::create('XenForo_Model_Node')->updateNestedSetInfo();
    		XenForo_Model::create('XenForo_Model_Permission')->rebuildPermissionCache();
    		return true;
    	}
    
    	$next = 0;
    	$last = 0;
    	$total = 0;
    	$position = 0;
    	$userIdMap = $model->getUserIdsMapFromArray($attachments, 'userid');
    
    	foreach ($attachments AS $attachment)
    	{
    		$next = $attachment['groupid'];
    
    		if ($last <> $next)
    		{
    			$last = $next;
    			$userid = $model->mapUserId($attachment['userid'], 0);
    			$username = $this->_convertToUtf8($attachment['username'], true);
    			$username = $this->mbTrim($username,50,$userid);
    
    			$import = array(
    					'title' => $this->_convertToUtf8('Group : '.$attachment['title'], true),
    					'node_id' => $options['forumid'],
    					'user_id' => $userid,
    					'username' => $username,
    					'discussion_open' => 1,
    					'post_date' => $attachment['dateline'],
    					'reply_count' => 0,
    					'view_count' => 0,
    					'sticky' => 0,
    					'discussion_state' => $attachment['state'] == 'public' ? 'visible' : 'moderated',
    					'last_post_date' => $attachment['dateline'],
    					'last_post_username' => $username
    			);
    
    			$threadid = $model->importThread(0, $import);
    
    			$position = 0;
    			$acount = $options['alimit'];
    			$model->logImportData('groupimage', $attachment['groupid'], $threadid);
    		}
    
    		if ($acount >= $options['alimit'])
    		{
    			if ($position != 0)
    			{
    				$this->_db->query('
    						UPDATE xf_thread
    						SET reply_count = reply_count + 1
    						WHERE thread_id = ?
    					', $threadid);
    			}
    
    			$acount = 0;
    			$position++;
    
    			$import = array(
    					'thread_id' => $threadid,
    					'user_id' => $userid,
    					'username' => $username,
    					'post_date' => $attachment['dateline']+$position,
    					'message' => "Group Images $position \n\n",
    					'attach_count' => 0,
    					'ip' => '0.0.0.0',
    					'position' => $position,
    					'message_state' => 'visible'
    			);
    
    			$postid = $model->importPost(0, $import);
    		}
    
    		$attachment['userid'] = $userid;
    		$attachment['postid'] = $postid;
    
    		unset($attachment['message']);
    		$success = $this->_makeAttachment($options, $attachment);
    
    		if ($success)
    		{
    			$total++;
    			$acount++;
    		}
    	}
    
    	$options['processed'] += $total;
    	$this->_session->incrementStepImportTotal($total);
    
    	return array($next, $options, $this->_getProgressOutput($options['processed'], $options['rows']));
    }
    
    public function stepBlogs($start, array $options)
    {
    	$options = array_merge(array(
    			'max' => false,
    			'limit' => 100,
    			'processed' => 0,
    			'postLimit' => 500,
    			'postDateStart' => 0,
    	), $options);
    
    	$sDb = $this->_sourceDb;
    	$prefix = $this->_prefix;
    	$model = $this->_importModel;
    
    	$active = $sDb->fetchOne('
    			SELECT active
    			FROM ' . $prefix . 'product
    			WHERE productid = \'vbblog\'
    		');
    
    	if (!$active)
    	{
    		return true; // Blogs not installed or active.
    	}
    
    	if ($options['max'] === false)
    	{
    		$data = $sDb->fetchRow('
    				SELECT MAX(blogid) AS max, COUNT(blogid) AS rows
    				FROM ' . $prefix . 'blog
    			');
    
    		$options = array_merge($options,$data);
    			
    		if ($options['max'])
    		{
    			// Get board name
    			$board = $sDb->fetchOne('
    					SELECT value
    					FROM ' . $prefix . 'setting
    					WHERE varname = \'bbtitle\'
    				');
    
    			XenForo_Db::beginTransaction();
    			$options['forumid'] = $this->buildImportForum($this->_convertToUtf8('Blogs ('.$board.')', true), 920);
    			XenForo_Db::commit();
    		}
    
    		return array(0, $options, "Processing Blogs ...");
    	}
    
    	$threads = $sDb->fetchAll($sDb->limit('
    				SELECT *
    				FROM ' . $prefix . 'blog AS blog
    				INNER JOIN ' . $prefix . 'blog_text AS blog_text 
    				ON (blog.firstblogtextid = blog_text.blogtextid)
    				WHERE blog.blogid >= ' . $sDb->quote($start) . '
    				ORDER BY blog.blogid
    			', $options['limit']
    	));
    
    	if (!$threads)
    	{
    		// Rebuild the permission cache
    		XenForo_Model::create('XenForo_Model_Node')->updateNestedSetInfo();
    		XenForo_Model::create('XenForo_Model_Permission')->rebuildPermissionCache();
    		return true; // All done.
    	}
    
    	$next = 0;
    	$total = 0;
    	$totalPosts = 0;
    
    	XenForo_Db::beginTransaction();
    
    	foreach ($threads AS $thread)
    	{
    		if (trim($thread['username']) === '' || trim($thread['title']) === '')
    		{
    			continue;
    		}
    
    		$postDateStart = $options['postDateStart'];
    
    		$options['postDateStart'] = 0;
    		$next = $thread['blogid'] + 1;
    
    		$maxPosts = $options['postLimit'] - $totalPosts;
    
    		$posts = $sDb->fetchAll($sDb->limit('
    					SELECT blog_text.* 
    					FROM ' . $prefix . 'blog_text AS blog_text
    					INNER JOIN ' . $prefix . 'user AS user USING (userid)
    					WHERE blog_text.blogid = ' . $sDb->quote($thread['blogid']) . '
    					AND blog_text.dateline > ' . $sDb->quote($postDateStart) . '
    					ORDER BY blog_text.dateline
    				', $maxPosts
    		));
    
    		if (!$posts)
    		{
    			if ($postDateStart)
    			{
    				$total++;
    			}
    			continue;
    		}
    
    		if ($postDateStart)
    		{
    			$threadId = $options['threadid'];
    			$position = $this->_db->fetchOne('
    					SELECT MAX(position)
    					FROM xf_post
    					WHERE thread_id = ?
    				', $threadId);
    		}
    		else
    		{
    			if (!$options['forumid'])
    			{
    				continue;
    			}
    
    			$import = array(
    					'title' => $this->_convertToUtf8($thread['title'], true),
    					'node_id' => $options['forumid'],
    					'user_id' => $model->mapUserId($thread['userid'], 0),
    					'username' => $this->_convertToUtf8($thread['username'], true),
    					'discussion_open' => 1,
    					'post_date' => $thread['dateline'],
    					'reply_count' => $thread['comments_visible'],
    					'view_count' => $thread['views'] ,
    					'sticky' => 0,
    					'last_post_date' => $thread['lastcomment'],
    					'last_post_username' => $this->_convertToUtf8($thread['lastcommenter'], true)
    			);
    
    			$import['username'] = $this->mbTrim($import['username'],50,$import['user_id']);
    
    			switch ($thread['state'])
    			{
    				case 'moderation': $import['discussion_state'] = 'moderated'; break;
    				default: $import['discussion_state'] = $thread['state'];
    			}
    
    			$threadId = $model->importThread(0, $import);
    
    			$options['threadid'] = $threadId;
    			$model->logImportData('blog', $thread['blogid'], $threadId);
    
    			if (!$threadId)
    			{
    				continue;
    			}
    
    			$position = -1;
    		}
    
    		if ($threadId)
    		{
    			$first = true;
    			$userIdMap = $model->getUserIdsMapFromArray($posts, 'userid');
    
    			foreach ($posts AS $post)
    			{
    				if (trim($post['username']) === '' || trim($post['pagetext']) === '')
    				{
    					continue;
    				}
    
    				$post['pagetext'] = $this->_convertVideo($post['pagetext']);
    
    				$import = array(
    						'thread_id' => $threadId,
    						'user_id' => $this->_mapLookUp($userIdMap, $post['userid'], 0),
    						'username' => $this->_convertToUtf8($post['username'], true),
    						'post_date' => $post['dateline'],
    						'message' => $this->_convertToUtf8($post['pagetext']),
    						'attach_count' => 0,
    						'ip' => $post['ipaddress']
    				);
    
    				$import['username'] = $this->mbTrim($import['username'],50,$import['user_id']);
    
    				switch ($post['state'])
    				{
    					case 'deleted': $import['message_state'] = 'deleted'; $import['position'] = $position; break;
    					case 'moderation': $import['message_state'] = 'moderated'; $import['position'] = $position; break;
    					default: $import['message_state'] = 'visible'; $import['position'] = ++$position;
    				}
    
    				$postid = $model->importPost(0, $import);
    				$model->logImportData('blogcomment', $post['blogtextid'], $postid);
    
    				if ($first)
    				{
    					$first = false;
    					$model->logImportData('blogfirst', $thread['blogid'], $postid);
    				}
    
    				$options['postDateStart'] = $post['dateline'];
    				$totalPosts++;
    			}
    		}
    
    		if (count($posts) < $maxPosts)
    		{
    			$total++;
    			$options['postDateStart'] = 0;
    		}
    		else
    		{
    			break;
    		}
    	}
    
    	XenForo_Db::commit();
    
    	if ($options['postDateStart'])
    	{
    		$next--;
    	}
    
    	$options['processed'] += $total;
    	$this->_session->incrementStepImportTotal($total);
    
    	return array($next, $options, $this->_getProgressOutput($options['processed'], $options['rows']));
    }
    
    public function stepArticles($start, array $options)
    {
    	$options = array_merge(array(
    			'max' => false,
    			'cats' => 0,
    			'limit' => 50,
    			'processed' => 0,
    			'cmsTypes' => '',
    			'postLimit' => 500,
    			'postDateStart' => 0,
    	), $options);
    
    	$sDb = $this->_sourceDb;
    	$prefix = $this->_prefix;
    	$model = $this->_importModel;
    
    	$active = $sDb->fetchOne('
    			SELECT active
    			FROM ' . $prefix . 'product
    			WHERE productid = \'vbcms\'
    		');
    
    	if (!$active)
    	{
    		return true; // CMS not installed or active.
    	}
    
    	if ($options['max'] === false)
    	{
    		$content = $sDb->fetchAll('
    				SELECT contenttypeid, class
    				FROM ' . $prefix . 'contenttype
    				WHERE class IN (\'Article\',\'StaticPage\',\'PhpEval\',\'Section\')
    			');
    
    		$options['cmsTypes'] = Array('0');
    
    		foreach ($content AS $type)
    		{
    			if ($type['class'] <> 'Section')
    			{
    				$options['cmsTypes'][] = $type['contenttypeid'];
    			}
    			$options['cms'.$type['class']] = $type['contenttypeid'];
    		}
    
    		$options['cmsTypes'] = implode(',',$options['cmsTypes']);
    
    		$data = $sDb->fetchRow('
    				SELECT MAX(nodeid) AS max, COUNT(nodeid) AS rows
    				FROM ' . $prefix . 'cms_node
    				WHERE contenttypeid IN ('.$options['cmsTypes'].')
    			');
    
    		$options = array_merge($options,$data);
    
    		if ($options['max'])
    		{
    			// Get board name
    			$board = $sDb->fetchOne('
    					SELECT value
    					FROM ' . $prefix . 'setting
    					WHERE varname = \'bbtitle\'
    				');
    
    			XenForo_Db::beginTransaction();
    			$options['forumid'] = $this->buildImportForum($this->_convertToUtf8('CMS Articles ('.$board.')', true), 930);
    			XenForo_Db::commit();
    
    			$options['cats'] = 1;
    			return array(0, $options, "Processing Sections ...");
    		}
    	}
    
    	if ($options['cats'] == 1)
    	{
    		XenForo_Db::beginTransaction();
    
    		// Do all sections in one go
    		$cats = $sDb->fetchAll('
    				SELECT *
    				FROM ' . $prefix . 'cms_node AS cms_node 
    				INNER JOIN ' . $prefix . 'cms_nodeinfo AS cms_nodeinfo USING (nodeid)
    				WHERE contenttypeid = ' . $sDb->quote($options['cmsSection']) . '
    				ORDER BY nodeid
    			');
    
    		foreach ($cats AS $category)
    		{
    			$newnode = $this->buildImportForum($this->_convertToUtf8($category['title']), $category['nodeid'], $options['forumid']);
    			$model->logImportData('cmsnode', $category['nodeid'], $newnode);
    		}
    
    		XenForo_Db::commit();
    
    		$options['cats'] = 2;
    		return array(0, $options, "Processing Articles ...");
    	}
    
    	$threads = $sDb->fetchAll($sDb->limit('
    				SELECT *
    				FROM ' . $prefix . 'cms_node AS cms_node 
    				INNER JOIN ' . $prefix . 'user AS user USING (userid)
    				INNER JOIN ' . $prefix . 'cms_nodeinfo AS cms_nodeinfo USING (nodeid)
    				WHERE contenttypeid IN ('.$options['cmsTypes'].') 
    				AND cms_node.nodeid >= ' . $sDb->quote($start) . '
    				ORDER BY cms_node.nodeid
    			', $options['limit']
    	));
    
    	if (!$threads)
    	{
    		// Rebuild the permission cache
    		XenForo_Model::create('XenForo_Model_Node')->updateNestedSetInfo();
    		XenForo_Model::create('XenForo_Model_Permission')->rebuildPermissionCache();
    		return true; // All done.
    	}
    
    	$next = 0;
    	$total = 0;
    	$totalPosts = 0;
    
    	$nodeMap = $model->getImportContentMap('cmsnode');
    
    	XenForo_Db::beginTransaction();
    
    	foreach ($threads AS $thread)
    	{
    		if (trim($thread['username']) === '' || trim($thread['title']) === '')
    		{
    			continue;
    		}
    
    		$postDateStart = $options['postDateStart'];
    
    		$options['postDateStart'] = 0;
    		$next = $thread['nodeid'] + 1;
    
    		$maxPosts = $options['postLimit'] - $totalPosts;
    
    		$posts = $sDb->fetchAll($sDb->limit('
    					SELECT cms_node.nodeid, post.userid, post.username, 
    					post.dateline, post.ipaddress, post.parentid, cms_node.contenttypeid, 
    					post.pagetext as postmessage, cms_article.pagetext as articlemessage,
    					cms_nodeconfig.value as staticmessage, post.visible, post.postid
    					FROM ' . $prefix . 'cms_nodeinfo AS cms_nodeinfo 
    					INNER JOIN ' . $prefix . 'cms_node AS cms_node USING (nodeid)
    					INNER JOIN ' . $prefix . 'post AS post 
    					ON (cms_nodeinfo.associatedthreadid = post.threadid)
    					LEFT JOIN ' . $prefix . 'cms_article AS cms_article USING (contentid)
    					LEFT JOIN ' . $prefix . 'cms_nodeconfig AS cms_nodeconfig 
    					ON (cms_nodeconfig.nodeid = cms_node.nodeid AND cms_nodeconfig.name = "pagetext")
    					WHERE cms_node.nodeid = ' . $sDb->quote($thread['nodeid']) . '
    					ORDER BY cms_node.nodeid, post.dateline
    				', $maxPosts
    		));
    
    		if (!$posts)
    		{
    			if ($postDateStart)
    			{
    				$total++;
    			}
    			continue;
    		}
    
    		if ($postDateStart)
    		{
    			$threadId = $options['threadid'];
    			$position = $this->_db->fetchOne('
    					SELECT MAX(position)
    					FROM xf_post
    					WHERE thread_id = ?
    				', $threadId);
    		}
    		else
    		{
    			$forumid = $this->_mapLookUp($nodeMap, $thread['parentnode']);
    
    			if (!$forumid)
    			{
    				continue;
    			}
    
    			$import = array(
    					'title' => $this->_convertToUtf8($thread['title'], true),
    					'node_id' => $forumid,
    					'user_id' => $model->mapUserId($thread['userid'], 0),
    					'username' => $this->_convertToUtf8($thread['username'], true),
    					'discussion_open' => 1,
    					'post_date' => $thread['publishdate'],
    					'reply_count' => 0,
    					'view_count' => $thread['viewcount'] ,
    					'sticky' => 0,
    					'last_post_date' => $thread['lastupdated'],
    					'last_post_username' => 'vB4 Importer'
    			);
    
    			$import['username'] = $this->mbTrim($import['username'],50,$import['user_id']);
    
    			switch ($thread['setpublish'])
    			{
    				case 0: $import['discussion_state'] = 'moderated'; break;
    				default: $import['discussion_state'] = 'visible';
    			}
    
    			$threadId = $model->importThread(0, $import);
    
    			$options['threadid'] = $threadId;
    			$model->logImportData('cmsarticle', $thread['nodeid'], $threadId);
    
    			if (!$threadId)
    			{
    				continue;
    			}
    
    			$position = -1;
    		}
    
    		if ($threadId)
    		{
    			$first = true;
    			$userIdMap = $model->getUserIdsMapFromArray($posts, 'userid');
    
    			foreach ($posts AS $post)
    			{
    				if ($post['parentid'])
    				{
    					$post['pagetext'] = $post['postmessage'];
    				}
    				else
    				{
    					if ($post['contenttypeid'] == $options['cmsArticle'])
    					{
    						$post['pagetext'] = $post['articlemessage'];
    					}
    					else
    					{
    						$post['pagetext'] = $post['staticmessage'];
    					}
    				}
    
    				if (trim($post['username']) === '' || trim($post['pagetext']) === '')
    				{
    					continue;
    				}
    
    				$post['pagetext'] = $this->_convertVideo($post['pagetext']);
    
    				$import = array(
    						'thread_id' => $threadId,
    						'user_id' => $this->_mapLookUp($userIdMap, $post['userid'], 0),
    						'username' => $this->_convertToUtf8($post['username'], true),
    						'post_date' => $post['dateline'],
    						'message' => $this->_convertToUtf8($post['pagetext']),
    						'attach_count' => 0,
    						'ip' => $post['ipaddress']
    				);
    
    				$import['username'] = $this->mbTrim($import['username'],50,$import['user_id']);
    
    				switch ($post['visible'])
    				{
    					case 2: $import['message_state'] = 'deleted'; $import['position'] = $position; break;
    					case 0: $import['message_state'] = 'moderated'; $import['position'] = $position; break;
    					default: $import['message_state'] = 'visible'; $import['position'] = ++$position;
    				}
    
    				$postid = $model->importPost(0, $import);
    				$model->logImportData('cmscomment', $post['postid'], $postid);
    
    				if ($first)
    				{
    					$first = false;
    					$model->logImportData('cmsfirst', $thread['nodeid'], $postid);
    				}
    
    				$options['postDateStart'] = $post['dateline'];
    				$totalPosts++;
    			}
    		}
    
    		if (count($posts) < $maxPosts)
    		{
    			$total++;
    			$options['postDateStart'] = 0;
    		}
    		else
    		{
    			break;
    		}
    	}
    
    	XenForo_Db::commit();
    
    	if ($options['postDateStart'])
    	{
    		$next--;
    	}
    
    	$options['processed'] += $total;
    	$this->_session->incrementStepImportTotal($total);
    
    	return array($next, $options, $this->_getProgressOutput($options['processed'], $options['rows']));
    }
    
    protected function buildImportForum($title, $order, $category = 0)
    {
    	if (!$category)
    	{
    		$category = $this->getImportCategory();
    	}
    
    	$import = array(
    			'node_type_id' => 'Forum',
    			'title' => $title,
    			'description' => $title . ' Imported from vBulletin 4',
    			'display_order' => $order,
    			'parent_node_id' => $category,
    			'display_in_list' => 1, 
    			'discussion_count' => 0,
    			'message_count' => 0,
    			'last_post_date' => Time(),
    			'last_post_username' => 'vB 4 Importer',
    			'last_thread_title' => $title,
    	);
    
    	$nodeId = $this->_importModel->importForum(0, $import);
    
    	if ($nodeId)
    	{
    		$userGroups = $this->_db->fetchAll('
    				SELECT *
    				FROM xf_user_group
    			');
    
    		foreach ($userGroups AS $userGroup)
    		{
    			switch ($userGroup['user_group_id'])
    			{
    				case 2: // Registered
    					$newPerms = $this->_setForumNoAccess();
    					break;
    
    				case 3: // Admins
    					$newPerms = $this->_setForumReadWrite();
    					break;
    
    				case 4: // Moderators
    					$newPerms = $this->_setForumReadonly();
    					break;
    
    				default: // The rest
    					$newPerms = $this->_setForumDeny();
    			}
    
    			$this->_importModel->insertNodePermissionEntries($nodeId, $userGroup['user_group_id'], 0, $newPerms);
    		}
    	}
    	else
    	{
    		throw new XenForo_Exception($title . ' : Forum Creation Failed.');
    	}
    
    	return $nodeId ;
    }
    
    protected function getImportCategory()
    {
    	$nodeId = $this->_db->fetchOne('
    			SELECT node_id
    			FROM xf_node
    			WHERE node_type_id = "Category" AND node_name = "vB4 Importer"
    		');
    
    	if ($nodeId)
    	{
    		return $nodeId ;
    	}
    
    	return $this->buildImportCategory('vBulletin 4 Imported Data', 900);
    }
    
    protected function buildImportCategory($title, $order)
    {
    	$rev = self::getVersion();
    
    	$import = array(
    			'node_type_id' => 'Category',
    			'title' => $title,
    			'description' => 'Information Imported From vBulletin 4 Using Importer Version '.$rev,
    			'display_order' => $order,
    			'parent_node_id' => 0,
    			'node_name' => 'vB4 Importer',
    	);
    
    	$nodeId = $this->_importModel->importForum(0, $import);
    
    	if ($nodeId)
    	{
    		$userGroups = $this->_db->fetchAll('
    				SELECT *
    				FROM xf_user_group
    			');
    
    		foreach ($userGroups AS $userGroup)
    		{
    			switch ($userGroup['user_group_id'])
    			{
    				case 2: // Registered
    					$newPerms = $this->_setForumNoAccess();
    					break;
    
    				case 3: // Admins
    					$newPerms = $this->_setForumReadWrite();
    					break;
    
    				case 4: // Moderators
    					$newPerms = $this->_setForumReadonly();
    					break;
    
    				default: // The rest
    					$newPerms = $this->_setForumDeny();
    			}
    
    			$this->_importModel->insertNodePermissionEntries($nodeId, $userGroup['user_group_id'], 0, $newPerms);
    		}
    	}
    	else
    	{
    		throw new XenForo_Exception($title . ' : Category Creation Failed.');
    	}
    
    	return $nodeId ;
    }
    
    protected static function getVersion()
    {
    	return '1.' . substr('$Revision: 1.0 $',11,-2);
    }
    
    protected function _setForumDeny()
    {
    	$output = array();
    	$output['general']['viewNode'] = 'deny';
    	$output['forum']['postThread'] = 'reset';
    	$output['forum']['postReply'] = 'reset';
    	$output['forum']['editOwnPost'] = 'reset';
    	$output['forum']['deleteOwnPost'] = 'reset';
    	$output['forum']['deleteOwnThread'] = 'reset';
    	$output['forum']['viewAttachment'] = 'reset';
    	$output['forum']['uploadAttachment'] = 'reset';
    	$output['forum']['votePoll'] = 'reset';
    	return $output;
    }

    protected function _setForumReadWrite()
    {
    	$output = array();
    	$output['general']['viewNode'] = 'content_allow';
    	$output['forum']['postThread'] = 'content_allow';
    	$output['forum']['postReply'] = 'content_allow';
    	$output['forum']['editOwnPost'] = 'content_allow';
    	$output['forum']['deleteOwnPost'] = 'content_allow';
    	$output['forum']['deleteOwnThread'] = 'content_allow';
    	$output['forum']['viewAttachment'] = 'content_allow';
    	$output['forum']['uploadAttachment'] = 'content_allow';
    	$output['forum']['votePoll'] = 'content_allow';
    	return $output;
    }
    
    protected function _setForumReadonly()
    {
    	$output = array();
    	$output['general']['viewNode'] = 'content_allow';
    	$output['forum']['postThread'] = 'reset';
    	$output['forum']['postReply'] = 'reset';
    	$output['forum']['editOwnPost'] = 'unset';
    	$output['forum']['deleteOwnPost'] = 'unset';
    	$output['forum']['deleteOwnThread'] = 'unset';
    	$output['forum']['viewAttachment'] = 'content_allow';
    	$output['forum']['uploadAttachment'] = 'reset';
    	$output['forum']['votePoll'] = 'unset';
    	return $output;
    }
    
    protected function _setForumNoAccess()
    {
    	$output = array();
    	$output['general']['viewNode'] = 'reset';
    	$output['forum']['postThread'] = 'reset';
    	$output['forum']['postReply'] = 'reset';
    	$output['forum']['editOwnPost'] = 'reset';
    	$output['forum']['deleteOwnPost'] = 'reset';
    	$output['forum']['deleteOwnThread'] = 'reset';
    	$output['forum']['viewAttachment'] = 'reset';
    	$output['forum']['uploadAttachment'] = 'reset';
    	$output['forum']['votePoll'] = 'reset';
    	return $output;
    }
    
    protected function mbTrim($string, $len = 0, $id = 0)
    {
    	/*
    		Trims multi byte string.
    	Mainly used for usernames.
    	Adds prefix as well - if requested, and string exceeds limit.
    	*/
    	if (utf8_strlen($string) <= $len)
    	{
    		return $string;
    	}
    
    	if ($len < 2)
    	{
    		throw new XenForo_Exception('Multi-byte Trim Parameter Error : Len '.$len);
    	}
    
    	if ($id)
    	{
    		$string = $id.'#'.$string;
    	}
    
    	return utf8_substr($string,0,$len);
    }
    
    protected function _convertVideo($string)
    {
    	// Convert Video Tags to Media Tags.
    	$string = preg_replace('/\[video=vimeo\;(.+)(&.*)?"?\].*\[\/video\]/i', '[media=vimeo]$1[/media]' ,$string);
    	$string = preg_replace('/\[video=youtube\;(.+)(&.*)?"?\].*\[\/video\]/i', '[media=youtube]$1[/media]' ,$string);
    	$string = preg_replace('/\[video=facebook\;(.+)(&.*)?"?\].*\[\/video\]/i', '[media=facebook]$1[/media]' ,$string);
    
    	// Convert raw youtube links to media bbcode (Krochinzky)
    	$string = preg_replace('/\[url="?http:\/\/www\.youtube\.com\/watch\?v=(.+)(&.*)?"?\].*\[\/url\]/i', '[media=youtube]$1[/media]' ,$string);
    	$string = preg_replace('/\[url\]http:\/\/www\.youtube\.com\/watch\?v=(.+)(&.*)?\[\/url\]/i', '[media=youtube]$1[/media]' ,$string);
    
    	// Convert raw vimeo links to media bbcode (Krochinzky)
    	$string = preg_replace('/\[url\]http:\/\/(www\.)?vimeo\.com\/(\d+)\[\/url\]/i', '[media=vimeo]$2[/media]' ,$string);
    	$string = preg_replace('/\[url="?http:\/\/(www\.)?vimeo\.com\/(\d+)"?](.*)\[\/url\]/i', '[media=vimeo]$2[/media]' ,$string);
    
    	// Convert raw facebook links to media bbcode.
    	$string = preg_replace('/\[url="?http:\/\/www\.facebook\.com\/video\/video\.php\?v=(.+)(&.*)?"?\].*\[\/url\]/i', '[media=facebook]$1[/media]' ,$string);
    	$string = preg_replace('/\[url\]http:\/\/www\.facebook\.com\/video\/video\.php\?v=(.+)(&.*)?\[\/url\]/i', '[media=facebook]$1[/media]' ,$string);
    
    	return trim($string);
    }

    
}